from yaetos.etl_utils import ETL_Base, Commandliner


class Job(ETL_Base):

    def transform(self, table_to_copy):
        if self.jargs.output.get('df_type', 'spark') == 'spark':
            table_to_copy.cache()
            if table_to_copy.count() < 500000:
                table_to_copy = table_to_copy.repartition(1)
        return table_to_copy


if __name__ == "__main__":
    args = {'job_param_file': 'conf/jobs_metadata.yml'}
    Commandliner(Job, **args)



import time
import botocore
from yaetos.deploy_utils import terminate
from yaetos.logger import setup_logging
logger = setup_logging('Deploy')


class EMRer():
    # TODO: change function call later. Current approach is quickest way to isolate code but not clean.
    @staticmethod
    def run_direct(self):
        """Useful to run job on cluster without bothering with aws data pipeline. Also useful to add steps to existing cluster."""
        self.s3_ops(self.session)
        if self.deploy_args.get('push_secrets', False):
            self.push_secrets(creds_or_file=self.app_args['connection_file'])  # TODO: fix privileges to get creds in dev env

        # EMR ops
        c = self.session.client('emr')
        clusters = self.get_active_clusters(c)
        cluster = self.choose_cluster(clusters)
        new_cluster = cluster['id'] is None
        if new_cluster:
            logger.info("Starting new cluster")
            self.start_spark_cluster(c, self.emr_version)
            logger.info("cluster name: %s, and id: %s" % (self.pipeline_name, self.cluster_id))
            self.step_run_setup_scripts(c)
            try:
                self.step_run_setup_scripts(c)
            except botocore.exceptions.ClientError as e:
                self.describe_status(c)
                logger.error(f"botocore.exceptions.ClientError : {e}")
                raise
        else:
            logger.info("Reusing existing cluster, name: %s, and id: %s" % (cluster['name'], cluster['id']))
            self.cluster_id = cluster['id']
            self.step_run_setup_scripts(c)

        # Run job
        self.step_spark_submit(c, self.app_file, self.app_args)

        # Clean
        if new_cluster and not self.deploy_args.get('leave_on') and self.app_args.get('clean_post_run'):  # TODO: add clean_post_run in input options.
            logger.info("New cluster setup to be deleted after job finishes.")
            self.describe_status_until_terminated(c)
            s3 = self.session.resource('s3')
            self.remove_temp_files(s3)  # TODO: remove tmp files for existing clusters too but only tmp files for the job

    @staticmethod
    def get_active_clusters(self, c):
        response = c.list_clusters(
            ClusterStates=['STARTING', 'BOOTSTRAPPING', 'RUNNING', 'WAITING'],
        )
        clusters = [(ii + 1, item['Id'], item['Name']) for ii, item in enumerate(response['Clusters'])]
        return clusters

    @staticmethod
    def choose_cluster(self, clusters, cluster_id=None):
        if len(clusters) == 0:
            logger.info('No cluster found, will create a new one')
            return {'id': None,
                    'name': None}

        if cluster_id is not None:
            logger.info('Cluster_id set by user to {}'.format(cluster_id))
            return {'id': cluster_id,
                    'name': None}

        clusters.append((len(clusters) + 1, None, 'Create a new cluster'))
        print('Clusters found for AWS account "%s":' % (self.aws_setup))
        print('\n'.join(['[%s] %s' % (item[0], item[2]) for item in clusters]))
        answer = input('Your choice ? ')
        return {'id': clusters[int(answer) - 1][1],
                'name': clusters[int(answer) - 1][2]}

    @staticmethod
    def start_spark_cluster(self, c, emr_version):
        """
        :param c: EMR client
        :return:
        """
        instance_groups = [{
            'Name': 'EmrMaster',
            'InstanceRole': 'MASTER',
            'InstanceType': self.ec2_instance_master,
            'InstanceCount': 1,
        }]
        if self.emr_core_instances != 0:
            instance_groups += [{
                'Name': 'EmrCore',
                'InstanceRole': 'CORE',
                'InstanceType': self.ec2_instance_slaves,
                'InstanceCount': self.emr_core_instances,
            }]

        response = c.run_job_flow(
            Name=self.pipeline_name,
            LogUri="s3://{}/{}/manual_run_logs/".format(self.s3_bucket_logs, self.metadata_folder),
            ReleaseLabel=emr_version,
            Instances={
                'InstanceGroups': instance_groups,
                'Ec2KeyName': self.ec2_key_name,
                'KeepJobFlowAliveWhenNoSteps': self.deploy_args.get('leave_on', False),
                'Ec2SubnetId': self.ec2_subnet_id,
                # 'AdditionalMasterSecurityGroups': self.extra_security_gp,  # TODO : make optional in future. "[self.extra_security_gp] if self.extra_security_gp else []" doesn't work.
            },
            Applications=self.emr_applications,  # should be at a minimum [{'Name': 'Hadoop'}, {'Name': 'Spark'}],
            Configurations=[
                {  # Section to force python3 since emr-5.x uses python2 by default.
                    "Classification": "spark-env",
                    "Configurations": [{
                        "Classification": "export",
                        "Properties": {"PYSPARK_PYTHON": "/usr/bin/python3"}
                    }]
                },
                # { # Section to add jars (redshift...), not used for now, since passed in spark-submit args.
                # "Classification": "spark-defaults",
                # "Properties": { "spark.jars": ["/home/hadoop/redshift_tbd.jar"], "spark.driver.memory": "40G", "maximizeResourceAllocation": "true"},
                # }
            ],
            JobFlowRole=self.emr_ec2_role,
            ServiceRole=self.emr_role,
            VisibleToAllUsers=True,
            BootstrapActions=[{
                'Name': 'setup_nodes',
                'ScriptBootstrapAction': {
                    'Path': 's3n://{}/setup_nodes.sh'.format(self.package_path_with_bucket),
                    'Args': []
                }
            }],
        )
        # Process response to determine if Spark cluster was started, and if so, the JobFlowId of the cluster
        response_code = response['ResponseMetadata']['HTTPStatusCode']
        if response_code == 200:
            self.cluster_id = response['JobFlowId']
        else:
            terminate("Could not create EMR cluster (status code {})".format(response_code))

        logger.info("Created Spark EMR cluster ({}) with cluster_id {}".format(emr_version, self.cluster_id))

    @staticmethod
    def describe_status_until_terminated(self, c):
        """
        :param c:
        :return:
        """
        logger.info('Waiting for job to finish on cluster')
        stop = False
        while stop is False:
            description = c.describe_cluster(ClusterId=self.cluster_id)
            state = description['Cluster']['Status']['State']
            if state == 'TERMINATED' or state == 'TERMINATED_WITH_ERRORS':
                stop = True
                logger.info('Job is finished')
            logger.info('Cluster state:' + state)
            time.sleep(30)  # Prevent ThrottlingException by limiting number of requests

    @staticmethod
    def describe_status(self, c):
        description = c.describe_cluster(ClusterId=self.cluster_id)
        logger.info(f'Cluster description: {description}')

    @staticmethod
    def step_run_setup_scripts(self, c):
        """
        :param c:
        :return:
        """
        response = c.add_job_flow_steps(
            JobFlowId=self.cluster_id,
            Steps=[{
                'Name': 'Run Setup',
                'ActionOnFailure': 'CONTINUE',
                'HadoopJarStep': {
                    'Jar': f's3://{self.s3_region}.elasticmapreduce/libs/script-runner/script-runner.jar',
                    'Args': [
                        "s3://{}/setup_master.sh".format(self.package_path_with_bucket),
                        "s3://{}".format(self.package_path_with_bucket),
                    ]
                }
            }]
        )
        response_code = response['ResponseMetadata']['HTTPStatusCode']
        if response_code == 200:
            logger.debug(f"Added step 'run setup', using s3://{self.package_path_with_bucket}/setup_master.sh")
        else:
            raise Exception("Step couldn't be added")
        time.sleep(1)  # Prevent ThrottlingException

    @staticmethod
    def step_spark_submit(self, c, app_file, app_args):
        """
        :param c:
        :return:
        """
        cmd_runner_args = self.get_spark_submit_args(app_file, app_args)

        response = c.add_job_flow_steps(
            JobFlowId=self.cluster_id,
            Steps=[{
                'Name': 'Spark Application',
                'ActionOnFailure': 'CONTINUE',
                'HadoopJarStep': {
                    'Jar': 'command-runner.jar',
                    'Args': cmd_runner_args
                }
            }]
        )
        response_code = response['ResponseMetadata']['HTTPStatusCode']
        if response_code == 200:
            logger.info("Added step 'spark-submit' with command line '{}'".format(' '.join(cmd_runner_args)))
        else:
            raise Exception("Step couldn't be added")
        time.sleep(1)  # Prevent ThrottlingException



import pandas as pd
from ConfigParser import ConfigParser
import os

os.environ[
    "NLS_LANG"
] = "AMERICAN_AMERICA.AL32UTF8"  # to be placed before import cx_Oracle, necessary to use queries with unicode.
# see https://github.com/oracle/python-cx_Oracle/issues/36 and https://stackoverflow.com/questions/49174710/how-to-handle-unicode-data-in-cx-oracle-and-python-2-7
import cx_Oracle


def connect(db, connection_type="sqlalchemy"):
    config = ConfigParser()
    config.read(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), "credentials.cfg")
    )

    params = {
        "user": config.get(db, "user"),
        "password": config.get(db, "password"),
        "host": config.get(db, "host"),
        "port": config.get(db, "port"),
        "service": config.get(db, "service"),
    }

    if connection_type == "oracle":
        # Basic connection, works for read only queries.
        connection_str = "{user}/{password}@{host}:{port}/{service}".format(**params)
        return cx_Oracle.connect(connection_str, encoding="UTF-8", nencoding="UTF-8")
    elif connection_type == "sqlalchemy":
        # Required to create table in oracle using df.to_sql()
        from sqlalchemy import create_engine

        connection_str = "oracle+cx_oracle://{user}:{password}@{host}:{port}/?service_name={service}".format(
            **params
        )  # Ex: 'oracle+cx_oracle://scott:tiger@host:1521/?service_name=hr'
        conn = create_engine(connection_str, encoding="UTF8")
        return conn
    else:
        raise ValueError("connection_type is not accepted: %s" % connection_type)


def query(query_str, **connect_args):
    """
    Run an sql query in oracle and return output to pandas table (dataframe).
    """
    connection = connect(**connect_args)

    if (
        connection.__class__.__doc__ is not None and connection.__class__.__doc__.__contains__("sqlalchemy")
    ):  # hasattr(connection, '__class__')
        df = pd.read_sql(query_str, connection, coerce_float=True)
        return df
    else:
        # TODO: fix rare pb "ValueError: invalid literal for int() with base 10: 'xxx.xx'"
        cursor = connection.cursor()
        try:
            cursor.execute(query_str)
        except cx_Oracle.DatabaseError as e:
            raise (
                cx_Oracle.DatabaseError,
                'Failed running query, received error "%s"' % e,
            )

        columns = [col[0] for col in cursor.description]
        df = pd.DataFrame.from_records(cursor.fetchall(), columns=columns)
        connection.close()
        return df


if __name__ == "__main__":
    df = query(
        "SELECT * FROM all_tables", db="name_of_connection_from_credentials_file"
    )
    print(df)



import pandas as pd
from configparser import ConfigParser
import os
from sqlalchemy import create_engine


def connect(db, creds_or_file='conf/connections.cfg'):
    config = ConfigParser()
    if isinstance(creds_or_file, str):
        assert os.path.isfile(creds_or_file)
        config.read(creds_or_file)
    else:
        config = creds_or_file

    params = {
        'user': config.get(db, 'user'),
        'password': config.get(db, 'password'),
        'host': config.get(db, 'host'),
        'port': config.get(db, 'port'),
        'service': config.get(db, 'service')
    }

    connection_str = "redshift+psycopg2://{user}:{password}@{host}:{port}/{service}".format(**params)
    conn = create_engine(connection_str, encoding='UTF8')
    return conn


def query(query_str, **connect_args):
    connection = connect(**connect_args)
    df = pd.read_sql(query_str, connection)
    return df


if __name__ == "__main__":
    df = query('SELECT * FROM yourtable', db='name_of_connection_from_credentials_file')
    print(df)



import yaml
import os
import subprocess
from yaetos.logger import setup_logging
logger = setup_logging('Job')


class Git_Config_Manager():

    FNAME = 'conf/git_config.yml'

    def get_config(self, mode, **kwargs):

        # Deal with multiple modes if any (Hacky. TODO: improve)
        modes = mode.split(',')
        if 'dev_local' in modes:
            config = self.get_config_from_git(kwargs['local_app_folder'])
            # For debug: self.save_yaml(config)
        elif 'dev_EMR' in modes or 'prod_EMR' in modes:  # TODO: remove dependency on 'dev_EMR' and 'prod_EMR'.
            config = self.get_config_from_file(kwargs['cluster_app_folder'])
        else:
            required_mode = ('dev_local', 'dev_EMR', 'prod_EMR')  # TODO: same
            raise Exception(f'Wrong mode, one of the mode should be in {required_mode}')
        return config

    def get_config_from_git(self, local_app_folder):

        # Current directory, i.e. yeatos job folder
        branch = subprocess.check_output(["git", "describe", '--all']).strip().decode('ascii')  # to get if dirty, add '--dirty'.
        last_commit = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('ascii')
        diffs = subprocess.check_output(['git', 'diff', 'HEAD']).strip().decode('ascii')
        is_dirty = True if diffs else False

        # Yaetos directory, i.e. framework folder. TODO: check how to handle when framework code is pulled from pip installed lib.
        branch_yaetos = subprocess.check_output(['git', 'describe', '--all'], cwd=local_app_folder).strip().decode('ascii')
        last_commit_yaetos = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=local_app_folder).strip().decode('ascii')
        diffs_yaetos = subprocess.check_output(['git', 'diff', 'HEAD'], cwd=local_app_folder).strip().decode('ascii')
        is_dirty_yaetos = True if diffs_yaetos else False

        config = {'branch_current': branch,
                  'last_commit_current': last_commit,
                  'diffs_current': diffs,
                  'is_dirty_current': is_dirty,
                  'branch_yaetos': branch_yaetos,
                  'last_commit_yaetos': last_commit_yaetos,
                  'diffs_yaetos': diffs_yaetos,
                  'is_dirty_yaetos': is_dirty_yaetos
                  }
        return config

    def is_git_controlled(self):
        out = os.system('git rev-parse')  # not using subprocess.check_output() to avoid crash if it fails.
        if out == 0:
            return True
        else:
            return False  # will send "fatal: not a git repository" or "git: command not found" to stderr

    def save_yaml(self, config):
        os.makedirs(os.path.dirname(self.FNAME), exist_ok=True)
        with open(self.FNAME, 'w') as file:
            yaml.dump(config, file)
        logger.info('Saved yml with git info: {}'.format(self.FNAME))

    def get_config_from_file(self, cluster_app_folder):
        """Meant to work in EMR"""
        fname = cluster_app_folder + self.FNAME
        if os.path.isfile(fname):
            with open(fname, 'r') as stream:
                yml = yaml.load(stream, Loader=yaml.FullLoader)
            return yml
        else:
            return False



"""Helper functions for clickhouse. Using postgres connector."""

from yaetos.logger import setup_logging
logger = setup_logging('Clickhouse')


def create_table(df, connection_profile, name_tb, schema, creds_or_file, is_incremental):
    """
    Creates table in Clickhouse, full drop or incremental drop, using spark connector. Implies pushing data to S3 first.
    """
    load_type = 'overwrite' if not is_incremental else 'append'
    db = creds_or_file[connection_profile]
    url = 'jdbc:postgresql://{host}/{service}'.format(host=db['host'], service=db['service'])
    dbtable = '{}.{}'.format(schema, name_tb)

    logger.info('Sending table "{}" to clickhouse in schema "{}", load type "{}", size "{}".'.format(name_tb, schema, load_type, df.count()))

    df.write \
        .format('jdbc') \
        .option('driver', "org.postgresql.Driver") \
        .option("url", url) \
        .option("user", db['user']) \
        .option("password", db['password']) \
        .option("dbtable", dbtable)\
        .mode(load_type) \
        .save()

    logger.info("Copied table to Clickhouse '{}.{}', using connection profile '{}'".format(schema, name_tb, connection_profile))



"""OBSOLETE. Will be deleted."""

from yaetos.etl_utils import ETL_Base
from datetime import datetime
from dateutil.relativedelta import relativedelta


class ETL_Daily_Incremental_Base(ETL_Base):
    def __init__(self, pre_jargs={}, jargs=None, loaded_inputs={}):
        super(ETL_Daily_Incremental_Base, self).__init__(pre_jargs, jargs, loaded_inputs)
        self.last_attempted_period = None

    def transform(self, **loaded_datasets):
        return self.get_transform_inc_from_last_days(**loaded_datasets)

    def transform_inc(self, period, **loaded_datasets):
        """ The function that needs to be overriden by each specific job."""
        raise NotImplementedError

    def get_transform_inc_from_last_days(self, **loaded_datasets):
        """ Incremental assumes last available month from the previous output was fully loaded."""
        first_day = self.jargs.merged_args['first_day']
        if not self.last_attempted_period:
            previous_output_max_timestamp = self.get_previous_output_max_timestamp()
            self.last_attempted_period = previous_output_max_timestamp.strftime("%Y-%m-%d") if previous_output_max_timestamp else first_day  # TODO: if get_output_max_timestamp()=None, means new build, so should delete instance in DBs.

        periods = self.get_last_output_to_last_day(self.last_attempted_period, first_day)
        if len(periods) == 0:
            self.logger.info('Output up to date. Nothing to run. last processed period={} and last period from now={}'.format(self.last_attempted_period, self.get_last_day()))
            self.final_inc = True
            return None

        self.logger.info('Periods remaining to load: {}'.format(periods))
        period = periods[0]
        self.logger.info('Period to be loaded in this run: {}'.format(period))
        self.final_inc = period == periods[-1]
        self.last_attempted_period = period
        self.jargs.merged_args['file_tag'] = period

        df = self.transform_inc(period, **loaded_datasets)
        return df

    @staticmethod
    def get_last_day():
        last_day_dt = datetime.utcnow() + relativedelta(days=-1)
        last_day = last_day_dt.strftime("%Y-%m-%d")
        return last_day

    @staticmethod
    def get_start_to_last_day(first_day):
        now = datetime.utcnow()
        start = datetime.strptime(first_day, "%Y-%m-%d")
        delta = now - start
        number_days = delta.days

        periods = []
        iter_days = start
        for item in range(number_days):
            periods.append(iter_days.strftime("%Y-%m-%d"))
            iter_days = iter_days + relativedelta(days=+1)
        return periods

    def get_last_output_to_last_day(self, last_output, first_day):
        periods = self.get_start_to_last_day(first_day)
        return [item for item in periods if item > last_output]



"""
File needed to run jobs from the yaml manifest (jobs_metadata.yml)
Usage: $ python jobs/generic/launcher.py --job_name=some_job_name_from_manifest
"""
from yaetos.etl_utils import Commandliner


args = {
    'job_param_file': 'conf/jobs_metadata.yml',
    'launcher_file': 'jobs/generic/launcher.py',
}
Commandliner(Job=None, **args)



from yaetos.etl_utils import ETL_Base, Commandliner


class Job(ETL_Base):
    def transform(self):
        return None


if __name__ == "__main__":
    args = {'job_param_file': 'conf/jobs_metadata.yml'}
    Commandliner(Job, **args)



from pyhive import hive
import pandas as pd
from ConfigParser import ConfigParser
import os


def connect(db):
    config = ConfigParser()
    config.read(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'credentials.cfg'))

    params = {
        'user': config.get(db, 'user'),
        'host': config.get(db, 'host'),
        'port': config.get(db, 'port') if config.has_option(db, 'port') else "10000",
    }
    return hive.Connection(host=params['host'], port=params['port'], username=params['user'])


def query(query_str, **kwarg):
    """
    Run an sql query in oracle and return output to pandas table (dataframe).
    """
    connection = connect(**kwarg)
    cursor = connection.cursor()

    try:
        cursor.execute(query_str)
    except Exception as err:
        raise Exception('Failed running hive query, with error: {}'.format(err))

    columns = [col[0] for col in cursor.description]
    df = pd.DataFrame.from_records(cursor.fetchall(), columns=columns)
    connection.close()
    return df


if __name__ == "__main__":
    df = query("SHOW TABLES", db='name_of_connection_from_credentials_file')
    print(df)



import pandas as pd
import numpy as np
import hashlib


def compare_dfs_exact(df1, df2):
    print('Looking into diffs between previous and new dataset.')
    try:
        hash1 = hashlib.sha256(pd.util.hash_pandas_object(df1, index=True).values).hexdigest()
        hash2 = hashlib.sha256(pd.util.hash_pandas_object(df2, index=True).values).hexdigest()
        is_identical = hash1 == hash2
    except Exception:
        print("Diff computation failed so assuming files are not similar.")
        return False

    if is_identical:
        print('dfs are the same.')
        return True

    print('dfs are different.')

    diff_columns = list(set(df1.columns) - set(df2.columns))
    if diff_columns:
        print('diff_columns: ', diff_columns)
        return False

    print('columns are the same.')
    return False


def compare_dfs_fuzzy(df1, pks1, compare1, df2, pks2, compare2, strip=True, filter_deltas=True, threshold=0.01):
    print('Length df1', len(df1), '. Unique values in each field :\n', df1[pks1].nunique())
    print('Length df2', len(df2), '. Unique values in each field :\n', df2[pks2].nunique())

    if strip:
        df1 = df1[pks1 + compare1]
        df2 = df2[pks2 + compare2]

    # Join datasets
    df_joined = pd.merge(left=df1, right=df2, how='outer', left_on=pks1, right_on=pks2, indicator=True, suffixes=('_1', '_2'))
    print('Length df_joined (before filtering if any)', len(df_joined))
    df_joined['_no_deltas'] = df_joined['_merge'].apply(lambda cell: True if cell == 'both' else False)  # init to True for 'both'. To be updated later.

    def check_delta(row):
        if pd.isna(row[item1]) and pd.isna(row[item2]):
            return 0
        elif pd.isna(row[item1]) or pd.isna(row[item2]):
            return 100
        elif float(row[item1]) == 0 and float(row[item2]) == 0:
            return 0
        elif float(row[item1]) == 0 or float(row[item2]) == 0:
            return 100
        else:
            return np.abs(np.divide((row[item1] - row[item2]), float(row[item1]))) * 100

    # Check deltas (numerical values)
    np.seterr(divide='ignore')  # to handle the division by 0 in divide().
    for ii in range(len(compare1)):
        item1 = compare1[ii]
        item2 = compare2[ii]
        if item1 == item2:  # check necessary for next step. See comment below.
            item1 = item1 + '_1'
            item2 = item2 + '_2'
        is_numeric_1 = pd.api.types.is_numeric_dtype(df_joined[item1])
        is_numeric_2 = pd.api.types.is_numeric_dtype(df_joined[item2])
        is_str_1 = pd.api.types.is_string_dtype(df_joined[item1])
        is_str_2 = pd.api.types.is_string_dtype(df_joined[item2])
        print(f'Starting to compare column ({item1} and {item2}), with types ({df_joined[item1].dtype} and {df_joined[item2].dtype})')
        if is_numeric_1 and is_numeric_2:
            try:
                df_joined['_delta_' + item1] = df_joined.apply(lambda row: (row[item1] if not pd.isna(row[item1]) else 0.0) - (row[item2] if not pd.isna(row[item2]) else 0.0), axis=1)
                df_joined['_delta_' + item1 + '_%'] = df_joined.apply(check_delta, axis=1)
                df_joined['_no_deltas'] = df_joined.apply(lambda row: row['_no_deltas'] is True and row['_delta_' + item1 + '_%'] < threshold, axis=1)
                print(f"Column summary, all_equal = {df_joined['_delta_' + item1 + '_%'].apply(lambda cell: cell < threshold).all()}. within treshold of {threshold}")
            except Exception as err:
                raise Exception("Failed item={}, error: \n{}".format(item1, err))
        elif is_str_1 and is_str_2:
            df_joined['_delta_' + item1 + '_equal'] = df_joined.apply(lambda row: row[item1] == row[item2], axis=1)
            df_joined['_no_deltas'] = df_joined.apply(lambda row: row['_no_deltas'] is True and row['_delta_' + item1 + '_equal'], axis=1)
            print(f"Column summary, all_equal = {df_joined['_delta_' + item1 + '_equal'].all()}.")
        else:
            try:
                df_joined['_delta_' + item1 + '_equal'] = df_joined.apply(lambda row: row[item1] == row[item2], axis=1)
                df_joined['_no_deltas'] = df_joined.apply(lambda row: row['_no_deltas'] is True and row['_delta_' + item1 + '_equal'], axis=1)
                print(f"Column summary, all_equal = {df_joined['_delta_' + item1 + '_equal'].all()}.")
            except Exception:
                df_joined['_no_deltas'] = df_joined.apply(lambda row: row['_no_deltas'] is False, axis=1)
            print(f'Column summary, all_equal = False. The columns to compare (i.e. {item1} and {item2}) have mismatched types, or are not numerical nor strings.')

    np.seterr(divide='raise')
    if filter_deltas:
        df_joined = df_joined[df_joined.apply(lambda row: row['_no_deltas'] is False, axis=1)].reset_index()
        print('Length df_joined (post filtering)', len(df_joined))
    df_joined.sort_values(by=['_merge'] + pks1, inplace=True)
    return df_joined



from ConfigParser import ConfigParser
import os


def get_user_pass(cred_profile, fpath=None):
    config = ConfigParser()
    if not fpath:
        fpath = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'api_creds.cfg')
    config.read(fpath)
    user = config.get(cred_profile, 'user')
    password = config.get(cred_profile, 'password')
    return user, password


def get_token(cred_profile, fpath=None):
    config = ConfigParser()
    if not fpath:
        fpath = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'api_creds.cfg')
    config.read(fpath)
    return config.get(cred_profile, 'token')



from yaetos.etl_utils import ETL_Base, Commandliner, Job_Args_Parser, Job_Yml_Parser


class Job(ETL_Base):
    """To run/deploy sql jobs, using --sql_file arg. Class reused by sql_pandas_job.sql"""

    def set_jargs(self, pre_jargs, loaded_inputs={}):
        # Function called only if running the job directly, i.e. "python jobs/generic/sql_spark_job.py --sql_file=jobs/some_job.sql",
        # ignored if running from "python jobs/generic/launcher.py --job_name=some_job.sql"
        # i.e. same behavior as other python jobs.
        sql_file = pre_jargs['cmd_args']['sql_file']
        if 'job_name' not in pre_jargs['cmd_args'].keys():
            job_name = Job_Yml_Parser.set_job_name_from_file(sql_file)
            pre_jargs['cmd_args']['job_name'] = job_name
        else:
            job_name = pre_jargs['job_args'].get('job_name')

        # Get all params from sql file and set as pre_jargs['job_args']
        query_str = self.read_sql_file(sql_file)
        job_args = self.get_params_from_sql(query_str)
        pre_jargs['job_args'].update(job_args)
        # 'job_name' param may be present in .sql file and so in pre_jargs['job_args'], but it will be overwritten by pre_jargs['cmd_args']['job_name'] if available.
        return Job_Args_Parser(defaults_args=pre_jargs['defaults_args'], yml_args=None, job_args=pre_jargs['job_args'], cmd_args=pre_jargs['cmd_args'], job_name=job_name, loaded_inputs=loaded_inputs)

    def transform(self, **ignored):
        sql = self.read_sql_file(self.jargs.sql_file)
        df = self.query(sql)
        if self.jargs.merged_args.get('repartition'):
            df = df.repartition(self.jargs.merged_args['repartition'])
        return df

    @staticmethod
    def read_sql_file(fname):
        fh = open(fname, 'r')
        sql = fh.read()
        fh.close()
        return sql

    @staticmethod
    def get_params_from_sql(sql):
        param_lines = [item.split('----')[2].split(':') for item in sql.split('\n') if (item.startswith('----param---- ') and len(item.split('----')) == 4)]
        params = {eval(key): eval(value) for key, value in param_lines}
        return params


if __name__ == "__main__":
    Commandliner(Job)



import pytest
try:
    from pyspark import SparkContext
    from pyspark.sql import SQLContext, SparkSession
except ModuleNotFoundError or ImportError as exc:
    print(f'Yaetos will work in pandas mode only or to push jobs to AWS, since pyspark is not found. Error "{exc}"')


@pytest.fixture(scope="session")
def sc(request):
    sc = SparkContext(appName='test')
    return sc


@pytest.fixture(scope="session")
def sc_sql(sc):
    return SQLContext(sc)


@pytest.fixture(scope="session")
def ss(sc):  # TODO: check if sc necessary here since not used downstream
    return SparkSession.builder.appName('test').getOrCreate()


@pytest.fixture
def get_pre_jargs():
    def internal_function(input_names=[], pre_jargs_over={}):
        """Requires loaded_inputs or pre_jargs_over"""
        inputs = {key: {'type': None} for key in input_names}
        pre_jargs = {
            'defaults_args': {
                'manage_git_info': False,
                'mode': 'dev_local',
                'deploy': 'none',
                'job_param_file': None,
                'output': {'type': None},
                'inputs': inputs},
            'job_args': {},
            'cmd_args': {}}
        if 'defaults_args' in pre_jargs_over.keys():
            pre_jargs['defaults_args'].update(pre_jargs_over['defaults_args'])
        else:
            pre_jargs.update(pre_jargs_over)
        return pre_jargs
    return internal_function


@pytest.fixture
def deploy_args():
    return {'aws_setup': 'dev',
            'aws_config_file': 'conf/aws_config.cfg.example',
            'mode': 'dev_EMR'}


@pytest.fixture
def app_args():
    return {'py_job': 'some/job.py', 'job_name': 'some_job_name'}



"""Helper functions for redshift."""

from libs.python_db_connectors.query_redshift import connect
from yaetos.logger import setup_logging
logger = setup_logging('Redshift')


def create_table(df, connection_profile, name_tb, schema, types, creds_or_file, is_incremental):
    """
    Creates table in redshift, full drop or incremental drop. Functional but very very slow.
    types should be of sqlalchemy type. Ex: types.Date(), types.Integer()
    """
    if_exist = 'replace' if not is_incremental else 'append'
    connection = connect(db=connection_profile, creds_or_file=creds_or_file)
    chunksize = 500000
    logger.info('Sending table "{}" to redshift in schema "{}", mode "{}", size "{}", and chunksize "{}".'.format(name_tb, schema, if_exist, len(df), chunksize))
    df.to_sql(name=name_tb, schema=schema, con=connection, if_exists=if_exist, dtype=types, index=False, chunksize=chunksize)
    # TODO: check df.to_sql above for long integers. Noticed long numbers where rounded.
    logger.info("Copied table to redshift '{}.{}', using connection profile '{}'".format(schema, name_tb, connection_profile))


if __name__ == '__main__':
    from sqlalchemy import types
    import pandas as pd
    data = [['aaa', 10], ['bbb', 12], ['ccc', 3]]
    df = pd.DataFrame(data, columns=['session_id', 'count_events'])
    types = {
        'session_id': types.VARCHAR(16),
        'count_events': types.Integer(),
    }
    connection_profile = 'some_connection_profile'
    name_tb = 'test_table'
    create_table(df, connection_profile, name_tb, types)



"""
Helper functions for pandas as engine.
"""
import pandas as pd
import glob
import json
import os
from pathlib import Path
from yaetos.logger import setup_logging
logger = setup_logging('Pandas')
try:
    import duckdb
    DUCKDB_SETUP = True
except ModuleNotFoundError or ImportError:
    logger.debug("DuckDB not found. Yaetos won't be able to run SQL jobs on top of pandas.")
    DUCKDB_SETUP = False


# --- loading files ----

def load_multiple_csvs(path, read_kwargs):
    # TODO: to be made obsolete once load_multiple_files works
    csv_files = glob.glob(os.path.join(path, "*.csv"))
    df = pd.concat((pd.read_csv(f, **read_kwargs) for f in csv_files))
    return df.reset_index(drop=True)


def load_multiple_files(path, globy='*.csv', read_func='read_csv', read_kwargs={}, add_file_fol=True):
    files = glob.glob(os.path.join(path, globy))  # removes local files
    files = [file for file in files if not os.path.basename(file).startswith('.')]  # remove sys files added by OS, ex .DS_store in mac.

    dfs = []
    for fi in files:
        df = load_df(fi, read_func, read_kwargs)
        if add_file_fol:
            df['_source'] = fi
        dfs.append(df)
    df = pd.concat(dfs) if len(dfs) >= 1 else pd.DataFrame()
    return df.reset_index(drop=True)


def load_csvs(path, read_kwargs):
    """Loading 1 csv or multiple depending on path"""
    # TODO: to be made obsolete once load_dfs works
    if path.endswith('.csv'):
        return pd.read_csv(path, **read_kwargs)
    elif path.endswith('/'):
        return load_multiple_csvs(path, read_kwargs)
    else:
        raise Exception("Path should end with '.csv' or '/'.".format())


def load_dfs(path, file_type='csv', globy=None, read_func='read_csv', read_kwargs={}):
    """Loading 1 file or multiple depending on path"""
    if globy:
        # TODO: improve check, make it usable with several files.
        matching_paths = glob.glob(globy)
        matches_pattern = os.path.normpath(path) in map(os.path.normpath, matching_paths)
        is_file = len(matching_paths) == 1

    if path.endswith(".{}".format(file_type)):  # one file and extension is explicite
        return load_df(path, read_func, read_kwargs)
    elif globy and matches_pattern and is_file:  # one file and extension is not explicite
        return load_df(path, read_func, read_kwargs)
    elif path.endswith('/'):  # multiple files.
        globy = globy or f'*.{file_type}'
        return load_multiple_files(path, globy, read_func, read_kwargs)
    else:  # case where file has no extension. TODO: make above more generic by using glob.
        raise Exception("Path should end with '.{}' or '/'.".format(file_type))


def load_df(path, read_func='read_csv', read_kwargs={}):
    """Loading 1 file or multiple depending on path"""
    if read_func != 'json_parser':
        func = getattr(pd, read_func)
        return func(path, **read_kwargs)
    else:
        with open(path, 'r') as file:
            data = json.load(file)

        df = pd.DataFrame(data['records'])
        return df


# --- saving files ----

def create_subfolders(path):
    """Creates subfolders if needed. Can take a path with a file."""
    output_dir = Path(path).parent
    output_dir.mkdir(parents=True, exist_ok=True)


def save_pandas_csv_local(df, path):
    # TODO: to be made obsolete once save_pandas_local works
    create_subfolders(path)
    df.to_csv(path, float_format='%.2f'.replace('.', ','))


def save_pandas_local(df, path, save_method='to_csv', save_kwargs={}):
    if isinstance(path, str):  # to deal with case when path is StringIO
        create_subfolders(path)
    func = getattr(df, save_method)
    func(path, **save_kwargs)


# --- other ----

def query_pandas(query_str, dfs):
    assert DUCKDB_SETUP is True
    con = duckdb.connect(database=':memory:')
    for key, value in dfs.items():
        con.register(key, value)
    df = con.execute(query_str).df()
    return df


def check_pk(df, pks):
    count = len(df)
    count_pk = len(df[pks].drop_duplicates())
    if count != count_pk:
        logger.error("Given fields ({}) are not PKs since not unique. count={}, count_pk={}".format(pks, count, count_pk))
        return False
    else:
        logger.info("Given fields ({}) are PKs (i.e. unique). count=count_pk={}".format(pks, count))
        return True


def add_created_at(pdf, start_dt):
    pdf['_created_at'] = start_dt
    return pdf



import json
import uuid
from datetime import datetime
from yaetos.logger import setup_logging
logger = setup_logging('Deploy')


class AWS_Data_Pipeliner():
    # TODO: change function call later. Current approach is quickest way to isolate code but not clean.
    @staticmethod
    def run_aws_data_pipeline(self):
        self.s3_ops(self.session)
        if self.deploy_args.get('push_secrets', False):
            self.push_secrets(creds_or_file=self.app_args['connection_file'])  # TODO: fix privileges to get creds in dev env

        # AWSDataPipeline ops
        client = self.session.client('datapipeline')
        self.deactivate_similar_pipelines(client, self.pipeline_name)
        pipe_id = self.create_data_pipeline(client)
        parameterValues = self.define_data_pipeline(client, pipe_id, self.emr_core_instances)
        self.activate_data_pipeline(client, pipe_id, parameterValues)

    @staticmethod
    def create_data_pipeline(self, client):
        unique_id = uuid.uuid1()
        create = client.create_pipeline(name=self.pipeline_name, uniqueId=str(unique_id))
        logger.debug('Pipeline created :' + str(create))

        pipe_id = create['pipelineId']  # format: 'df-0624751J5O10SBRYJJF'
        logger.info('Created pipeline with id ' + pipe_id)
        logger.debug('Pipeline description :' + str(client.describe_pipelines(pipelineIds=[pipe_id])))
        return pipe_id

    @staticmethod
    def define_data_pipeline(self, client, pipe_id, emr_core_instances):
        import awscli.customizations.datapipeline.translator as trans
        base = self.get_package_path()

        if emr_core_instances != 0:
            definition_file = base / 'yaetos/definition.json'  # see syntax in datapipeline-dg.pdf p285 # to add in there: /*"AdditionalMasterSecurityGroups": "#{}",  /* To add later to match EMR mode */
        else:
            definition_file = base / 'yaetos/definition_standalone_cluster.json'
            # TODO: have 1 json for both to avoid having to track duplication.

        definition = json.load(open(definition_file, 'r'))  # Note: Data Pipeline doesn't support emr-6.0.0 yet.

        pipelineObjects = trans.definition_to_api_objects(definition)
        parameterObjects = trans.definition_to_api_parameters(definition)
        parameterValues = trans.definition_to_parameter_values(definition)
        parameterValues = self.update_params(parameterValues)
        logger.debug(f'Filled pipeline with data from {definition_file}')

        response = client.put_pipeline_definition(
            pipelineId=pipe_id,
            pipelineObjects=pipelineObjects,
            parameterObjects=parameterObjects,
            parameterValues=parameterValues
        )
        logger.debug('put_pipeline_definition response: ' + str(response))
        return parameterValues

    @staticmethod
    def activate_data_pipeline(self, client, pipe_id, parameterValues):
        response = client.activate_pipeline(
            pipelineId=pipe_id,
            parameterValues=parameterValues,  # optional. If set, need to specify all params as per json.
            # startTimestamp=datetime(2018, 12, 1)  # optional
        )
        logger.debug('activate_pipeline response: ' + str(response))
        logger.info('Activated pipeline ' + pipe_id)

    @staticmethod
    def list_data_pipeline(self, client):
        out = client.list_pipelines(marker='')
        pipelines = out['pipelineIdList']
        while out['hasMoreResults'] is True:
            out = client.list_pipelines(marker=out['marker'])
            pipelines += out['pipelineIdList']
        return pipelines

    @staticmethod
    def deactivate_similar_pipelines(self, client, pipeline_id):
        pipelines = self.list_data_pipeline(client)
        for item in pipelines:
            job_name = self.get_job_name(item['name'])
            if job_name == self.app_args['job_name']:
                response = client.deactivate_pipeline(pipelineId=item['id'], cancelActive=True)
                response_code = response['ResponseMetadata']['HTTPStatusCode']
                if response_code == 200:
                    logger.info('Deactivated pipeline {}, {}, {}'.format(job_name, item['name'], item['id']))
                else:
                    raise Exception("Pipeline couldn't be deactivated. Error message: {}".format(response))

    @staticmethod
    def update_params(self, parameterValues):
        # TODO: check if easier/simpler to change values at the source json instead of a processed one.
        # Change key pair
        myScheduleType = {'EMR_Scheduled': 'cron', 'EMR_DataPipeTest': 'ONDEMAND'}[self.deploy_args.get('deploy')]
        myPeriod = self.deploy_args['frequency'] or '1 Day'
        if self.deploy_args['start_date'] and isinstance(self.deploy_args['start_date'], datetime):
            myStartDateTime = self.deploy_args['start_date'].strftime('%Y-%m-%dT%H:%M:%S')
        elif self.deploy_args['start_date'] and isinstance(self.deploy_args['start_date'], str):
            myStartDateTime = self.deploy_args['start_date'].format(today=datetime.today().strftime('%Y-%m-%d'))
        else:
            myStartDateTime = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')
        bootstrap = 's3://{}/setup_nodes.sh'.format(self.package_path_with_bucket)

        for ii, item in enumerate(parameterValues):
            if 'myEC2KeyPair' in item.values():
                parameterValues[ii] = {'id': u'myEC2KeyPair', 'stringValue': self.ec2_key_name}
            elif 'mySubnet' in item.values():
                parameterValues[ii] = {'id': u'mySubnet', 'stringValue': self.ec2_subnet_id}
            elif 'myPipelineLogUri' in item.values():
                parameterValues[ii] = {'id': u'myPipelineLogUri', 'stringValue': "s3://{}/{}/scheduled_run_logs/".format(self.s3_bucket_logs, self.metadata_folder)}
            elif 'myScheduleType' in item.values():
                parameterValues[ii] = {'id': u'myScheduleType', 'stringValue': myScheduleType}
            elif 'myPeriod' in item.values():
                parameterValues[ii] = {'id': u'myPeriod', 'stringValue': myPeriod}
            elif 'myStartDateTime' in item.values():
                parameterValues[ii] = {'id': u'myStartDateTime', 'stringValue': myStartDateTime}
            elif 'myBootstrapAction' in item.values():
                parameterValues[ii] = {'id': u'myBootstrapAction', 'stringValue': bootstrap}
            elif 'myTerminateAfter' in item.values():
                parameterValues[ii] = {'id': u'myTerminateAfter', 'stringValue': self.deploy_args.get('terminate_after', '180 Minutes')}
            elif 'myEMRReleaseLabel' in item.values():
                parameterValues[ii] = {'id': u'myEMRReleaseLabel', 'stringValue': self.emr_version}
            elif 'myMasterInstanceType' in item.values():
                parameterValues[ii] = {'id': u'myMasterInstanceType', 'stringValue': self.ec2_instance_master}
            elif 'myCoreInstanceCount' in item.values():
                parameterValues[ii] = {'id': u'myCoreInstanceCount', 'stringValue': str(self.emr_core_instances)}
            elif 'myCoreInstanceType' in item.values():
                parameterValues[ii] = {'id': u'myCoreInstanceType', 'stringValue': self.ec2_instance_slaves}

        # Change steps to include proper path
        setup_command = 's3://{s3_region}.elasticmapreduce/libs/script-runner/script-runner.jar,s3://{s3_tmp_path}/setup_master.sh,s3://{s3_tmp_path}'.format(s3_tmp_path=self.package_path_with_bucket, s3_region=self.s3_region)  # s3://elasticmapreduce/libs/script-runner/script-runner.jar,s3://bucket-tempo/ex1_frameworked_job.arthur_user1.20181129.231423/setup_master.sh,s3://bucket-tempo/ex1_frameworked_job.arthur_user1.20181129.231423/
        spark_submit_command = 'command-runner.jar,' + ','.join([item.replace(',', '\\\,') for item in self.get_spark_submit_args(self.app_file, self.app_args)])   # command-runner.jar,spark-submit,--py-files,/home/hadoop/app/scripts.zip,--packages=com.amazonaws:aws-java-sdk-pom:1.11.760\\\\,org.apache.hadoop:hadoop-aws:2.7.0,/home/hadoop/app/jobs/examples/ex1_frameworked_job.py,--storage=s3  # instructions about \\\ part: https://docs.aws.amazon.com/datapipeline/latest/DeveloperGuide/dp-object-emractivity.html

        commands = [setup_command, spark_submit_command]
        mm = 0
        for ii, item in enumerate(parameterValues):
            if 'myEmrStep' in item.values() and mm < 2:  # TODO: make more generic and cleaner
                parameterValues[ii] = {'id': u'myEmrStep', 'stringValue': commands[mm]}
                mm += 1

        logger.debug('parameterValues after changes: ' + str(parameterValues))
        return parameterValues



from jobs.generic.sql_pandas_job import Job
import pandas as pd


class Test_Job(object):
    def test_transform(self, get_pre_jargs):
        some_events = pd.DataFrame([
            {'session_id': 1234, 'action': 'searchResultPage', 'n_results': 1},
            {'session_id': 1234, 'action': 'searchResultPage', 'n_results': 1},
            {'session_id': 1235, 'action': 'searchResultPage', 'n_results': 1},
            {'session_id': 1236, 'action': 'other',            'n_results': 1},  # noqa: E241
        ])

        other_events = pd.DataFrame([
            {'session_id': 1234, 'other': 1},
            {'session_id': 1235, 'other': 1},
            {'session_id': 1236, 'other': 1},
        ])

        expected = [
            {'session_id': 1234, 'count_events': 2},
            {'session_id': 1235, 'count_events': 1},
            # only diff with ex1_framework_job is session_id being str instead of int.
        ]

        sql_file = 'jobs/examples/ex1_sql_job.sql'

        # Get actual
        loaded_inputs = {'some_events': some_events, 'other_events': other_events}
        pre_jargs = get_pre_jargs(loaded_inputs.keys())
        pre_jargs['cmd_args']['sql_file'] = sql_file
        job = Job(pre_jargs=pre_jargs)
        assert 'param_a' in job.jargs.merged_args.keys()
        actual = job.etl_no_io(sc=None, sc_sql=None, loaded_inputs=loaded_inputs)[0].to_dict(orient='records')

        # Compare
        assert actual == expected



"""Helper functions for athena."""
import boto3
import os
from configparser import ConfigParser
from yaetos.logger import setup_logging
logger = setup_logging('Athena')


def register_table_to_athena_catalog(types, name_tb, schema, output_info, args):
    description_statement = f"""COMMENT "{args['description']}" """ if args.get('description') else ''
    output_folder = output_info['path_expanded'].replace('s3a', 's3')

    types_str = ''
    ii_max = len(types) - 1
    for ii, (col_name, col_type) in enumerate(types.items()):
        types_str += f"`{col_name}` {col_type}"
        types_str += ", \n" if ii < ii_max else ''

    if output_info['type'] == 'csv':
        create_table_query = f"""
            CREATE EXTERNAL TABLE IF NOT EXISTS `{schema}`.`{name_tb}` (
            {types_str}
            ) {description_statement}
            ROW FORMAT SERDE 'org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe'
            WITH SERDEPROPERTIES ('field.delim' = ',')
            STORED AS INPUTFORMAT 'org.apache.hadoop.mapred.TextInputFormat' OUTPUTFORMAT 'org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat'
            LOCATION '{output_folder}'
            TBLPROPERTIES ('classification' = 'csv');
        """
        logger.info(f"Table registration with : \n{create_table_query} \n")
    elif output_info['type'] == 'parquet':
        create_table_query = f"""
            CREATE EXTERNAL TABLE IF NOT EXISTS `{schema}`.`{name_tb}` (
            {types_str}
            ) {description_statement}
            STORED AS PARQUET
            LOCATION '{output_folder}'
        """
        logger.info(f"Table registration with : \n{create_table_query} \n")
    else:
        raise Exception('Athena table registration not setup for other than csv files.')

    # Start the query execution
    if args.get('mode') and 'dev_local' in args['mode'].split(','):
        config = ConfigParser()
        assert os.path.isfile(args.get('aws_config_file'))
        config.read(args.get('aws_config_file'))
        region_name = config.get(args.get('aws_setup'), 's3_region')
    else:
        region_name = boto3.Session().region_name

    client = boto3.client('athena', region_name=region_name)
    response = client.start_query_execution(
        QueryString=create_table_query,
        QueryExecutionContext={'Database': schema},
        ResultConfiguration={'OutputLocation': args.get('athena_out')},
    )
    logger.info(f"Registered table to athena '{schema}.{name_tb}', pointing to {output_folder}, with QueryExecutionId: {response['QueryExecutionId']}.")
    # TODO: Check to support "is_incremental"


def register_table_to_glue_catalog(schema_list, name_tb, schema, output_info, args):
    output_folder = output_info['path_expanded'].replace('s3a', 's3')

    # Start the query execution
    if args.get('mode') == 'dev_local':
        config = ConfigParser()
        assert os.path.isfile(args.get('aws_config_file'))
        config.read(args.get('aws_config_file'))
        region_name = config.get(args.get('aws_setup'), 's3_region')
    else:
        region_name = boto3.Session().region_name

    glue_client = boto3.client('glue', region_name=region_name)

    database_name = schema
    table_name = name_tb
    logger.info(f"Registering table to athena '{schema}.{name_tb}', schema {schema_list}.")

    # Define the table input
    table_input = {
        'Name': table_name,
        'StorageDescriptor': {
            'Columns': schema_list,
            'Location': output_folder,
            'InputFormat': 'org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat',
            'OutputFormat': 'org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat',
            'Compressed': False,
            'SerdeInfo': {
                'SerializationLibrary': 'org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe',
                'Parameters': {
                    'serialization.format': '1'
                }
            },
        },
        'TableType': 'EXTERNAL_TABLE',
        'Parameters': {
            'classification': 'parquet'
        }
    }

    try:
        glue_client.delete_table(DatabaseName=database_name, Name=table_name)
        print("Existing table deleted.")
    except glue_client.exceptions.EntityNotFoundException:
        print("No table to delete.")

    response = glue_client.create_table(
        DatabaseName=database_name,
        TableInput=table_input
    )
    logger.info(f"Registered table to athena '{schema}.{name_tb}', pointing to {output_folder} with query response: {response}.")



from simple_salesforce import Salesforce
import pandas as pd
from configparser import ConfigParser
import os


def connect(creds_section, creds_or_file='conf/connections.cfg'):
    config = ConfigParser()
    if isinstance(creds_or_file, str):
        assert os.path.isfile(creds_or_file)
        config.read(creds_or_file)
    else:
        config = creds_or_file

    user = config.get(creds_section, 'user')
    pwd = config.get(creds_section, 'password')
    token = config.get(creds_section, 'token')
    domain = None if config.get(creds_section, 'domain') == 'production' else config.get(creds_section, 'domain')
    return Salesforce(username=user, password=pwd, security_token=token, domain=domain)


def query(query_str, **connect_args):
    sf = connect(**connect_args)
    resp = sf.query_all(query_str)
    rows = resp['records']
    for row in rows:
        row.__delitem__('attributes')
    df = pd.DataFrame.from_dict(rows)
    return df


if __name__ == "__main__":
    df = query('SELECT Account.Name FROM Account', creds_section='name_of_connection_from_credentials_file')
    print(df)



"""Helper functions for databases, based on sqlalchemy, mostly for oracle for now.
Some code to be run in worker nodes, so can't rely on libraries only on master (ex db connections)."""

import pandas as pd
from sqlalchemy import types as db_types
from pyspark.sql import types as spk_types
from datetime import datetime, date
from yaetos.logger import setup_logging
logger = setup_logging('DB_Utils')


def cast_rec(rec, output_types):
    new_rec = {}
    for field in output_types.keys():
        new_rec[field] = cast_value(rec[field], output_types[field], field)
    return new_rec


def cast_value(value, required_type, field_name):
    # TODO: make it less ugly.. or avoid using pandas to not require this.
    try:
        if isinstance(required_type, type(db_types.DATE())):
            if isinstance(value, str):
                return datetime.strptime(value, "%Y-%m-%d")  # assuming iso format
            elif isinstance(value, pd.Timestamp):  # == datetime
                return value.to_pydatetime().date()
            elif isinstance(value, date):
                return value
            elif pd.isnull(value):
                return None
            else:
                return required_type.python_type(value)
        if isinstance(required_type, type(db_types.DATETIME())):
            if isinstance(value, str):
                return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")  # assuming iso format
            elif isinstance(value, pd.Timestamp):
                return value.to_pydatetime()
            elif pd.isnull(value):
                return None
            else:
                return required_type.python_type(value)
        elif isinstance(required_type, type(db_types.VARCHAR())):
            return None if pd.isnull(value) else str(value)
        elif isinstance(required_type, type(db_types.INT())):
            return None if pd.isnull(value) else int(float(value))
        elif isinstance(required_type, type(db_types.BIGINT())):
            return None if pd.isnull(value) else int(value)
        elif isinstance(required_type, type(db_types.FLOAT())):
            return None if pd.isnull(value) else float(value)
        else:
            return required_type.python_type(value)
    except Exception as e:
        logger.error(u"cast_value issue: {}, {}, {}, {}, {}.".format(field_name, value, type(value), required_type, str(e)))
        return None


def cast_col(df, output_types):
    for field in df.columns:
        if isinstance(output_types[field], type(db_types.FLOAT())):
            df[field] = df[field].astype(float)
        if isinstance(output_types[field], type(db_types.DATE())):
            df[field] = df[field].astype('datetime64[ns]')
        # if isinstance(output_types[field], type(db_types.INT())):
        #     df[field] = df[field].astype(int)  -> this "if" leads to pb in prod when they are none value + handled properly in prod with string inputs being converted to int.
    return df


def get_spark_type(field, required_type):
    if isinstance(required_type, type(db_types.DATE())):
        return spk_types.StructField(field, spk_types.DateType(), True)
    elif isinstance(required_type, type(db_types.DATETIME())):
        return spk_types.StructField(field, spk_types.TimestampType(), True)
    elif isinstance(required_type, type(db_types.VARCHAR())):
        return spk_types.StructField(field, spk_types.StringType(), True)
    elif isinstance(required_type, type(db_types.INT())):
        return spk_types.StructField(field, spk_types.LongType(), True)  # db type enforced earlier than spark ones, so spark types needs to be less restrictive than spark ones so needs to choose LongType instead of IntegerType
    elif isinstance(required_type, type(db_types.FLOAT())):
        return spk_types.StructField(field, spk_types.FloatType(), True)
    elif isinstance(required_type, type(db_types.BOOLEAN())):
        return spk_types.StructField(field, spk_types.BooleanType(), True)
    else:
        raise Exception("Type not recognized, field={}, required_type={}".format(field, required_type))
    # BIGINT not usable here as not supported by Oracle.


def get_spark_types(output_types):
    spark_types = []
    for field, required_type in output_types.items():
        spark_type = get_spark_type(field, required_type)
        spark_types.append(spark_type)

    spark_schema = spk_types.StructType(spark_types)
    logger.info('spark_schema: {}'.format(spark_schema))
    return spark_schema


def pdf_to_sdf(df, output_types, sc, sc_sql):  # TODO: check suspicion that this leads to each node requiring loading all libs from this script.
    # can now be done with sdf = self.sc_sql.createDataFrame(pdf). TODO: check to deprecate pdf_to_sdf if createDataFrame(pdf) covers every cases.
    spark_schema = get_spark_types(output_types)
    missing_columns = set(df.columns) - set(output_types.keys())
    if missing_columns:
        logger.warning('Some fields from source pandas df will not be pushed to spark df (because of absence in output_types), check if need to be added: {}.'.format(missing_columns))

    recs = df.to_dict(orient='records')
    partitions = len(recs) / 1000
    partitions = partitions if partitions >= 1 else None
    # Push to spark. For easier testing of downstream casting (i.e. outside of spark): tmp = [cast_rec(row, output_types) for row in recs]
    rdd = sc.parallelize(recs, numSlices=partitions) \
            .map(lambda row: cast_rec(row, output_types))

    return sc_sql.createDataFrame(rdd, schema=spark_schema, verifySchema=True)


def pandas_types_to_hive_types(df, format='glue'):
    """
    Converts pandas DataFrame dtypes to Hive column types.

    :param df: pandas DataFrame
    :return: Dictionary or list of dictionaries of column names and their Hive data types
    """
    # TODO: add support for nested fields, just like in spark version
    type_mapping = {
        'object': 'STRING',
        'bool': 'BOOLEAN',
        'datetime64[ns]': 'TIMESTAMP',
        'timedelta[ns]': 'STRING',  # Hive does not have a direct equivalent
        'category': 'STRING',  # Hive has no direct category type; consider using STRING or a specific type based on the category

        # Integer types
        'int8': 'TINYINT',
        'int16': 'SMALLINT',
        'int32': 'INT',
        'int64': 'BIGINT',
        'uint8': 'SMALLINT',  # Hive does not have unsigned types, so use the next larger type
        'uint16': 'INT',
        'uint32': 'BIGINT',
        'uint64': 'BIGINT',  # Note: Hive BIGINT might not cover the full range of uint64

        # Floating types
        'float16': 'FLOAT',
        'float32': 'FLOAT',
        'float64': 'DOUBLE',

        # Special handling for decimals
        # This is a placeholder; actual handling should consider the specific precision and scale
        # 'decimal': 'DECIMAL',
    }
    if format == 'athena':
        hive_types = {}
        for column, dtype in df.dtypes.items():
            dtype_name = dtype.name
            hive_type = type_mapping.get(dtype_name, 'STRING')
            hive_types[column] = hive_type
        return hive_types
    elif format == 'glue':
        hive_types = []
        for column_name, dtype in df.dtypes.items():
            hive_type = type_mapping.get(str(dtype), 'string')
            hive_types.append({'Name': column_name, 'Type': hive_type})
        return hive_types


def spark_type_to_hive_type(data_type):
    """ Convert Spark data types to a detailed readable string format, handling nested structures. """
    from pyspark.sql.types import StructType, ArrayType, FloatType, DecimalType, TimestampType, IntegerType  # StructField, StringType,

    if isinstance(data_type, StructType):
        # Handle nested struct by recursively processing each field
        fields = [f"{field.name}: {spark_type_to_hive_type(field.dataType)}" for field in data_type.fields]
        return f"struct<{', '.join(fields)}>"
    elif isinstance(data_type, ArrayType):
        # Handle arrays by describing element types
        element_type = spark_type_to_hive_type(data_type.elementType)
        return f"array<{element_type}>"
    elif isinstance(data_type, TimestampType):
        return "timestamp"
    elif isinstance(data_type, IntegerType):
        return "int"
    elif isinstance(data_type, FloatType):
        return "float"
    elif isinstance(data_type, DecimalType):
        return f"decimal({data_type.precision},{data_type.scale})"
    else:
        # Fallback for other types with default string representation
        return data_type.simpleString()


def spark_types_to_hive_types(sdf):
    schema_list = [{"Name": field.name, "Type": spark_type_to_hive_type(field.dataType)} for field in sdf.schema]
    return schema_list



import os
from yaetos.logger import setup_logging
from yaetos.airflow_template_k8s import get_template_k8s
logger = setup_logging('Deploy')


class Kuberneter():
    # TODO: change function call later. Current approach is quickest way to isolate code but not clean.
    @staticmethod
    def run_direct_k8s(self):
        """Useful to run job on cluster on the spot, without going through scheduler."""
        self.local_file_ops()
        if self.deploy_args.get('push_secrets', False):
            self.push_secrets(creds_or_file=self.app_args['connection_file'])  # TODO: fix privileges to get creds in dev env

        logger.info("Sending spark-submit to k8s cluster")
        logger.info("To monitor progress:")
        logger.info(f"kubectl get pods -n {self.app_args['k8s_namespace']}")
        logger.info(f"kubectl logs your_spark_app_id_here -n {self.app_args['k8s_namespace']}")
        cmdline = self.get_spark_submit_args_k8s(self.app_file, self.app_args)
        self.launch_spark_submit_k8s(cmdline)
        logger.info("Spark submit finished, see results with:")
        logger.info(f"kubectl get pods -n {self.app_args['k8s_namespace']}")
        logger.info(f"kubectl logs your_spark_app_id_here -n {self.app_args['k8s_namespace']}")

    @staticmethod
    def get_spark_submit_args_k8s(app_file, app_args):
        """ app_file is launcher, might be py_job too, but may also be separate from py_job (ex python launcher.py --job_name=some_job_with_py_job)."""
        # TODO: need to unify with get_spark_submit_args(), to account for jar jobs, to use eu.Runner.create_spark_submit()
        # See spark-submit setup for k8s on docker_desktop.

        if app_args.get('spark_submit'):
            return app_args['spark_submit']

        # For k8s in AWS, for yaetos jobs.
        spark_submit_base = [
            'spark-submit',
            f'--master {app_args["k8s_url"]}',
            '--deploy-mode cluster',
            f'--name {app_args["k8s_name"]}',
            f'--conf spark.executor.instances={app_args["k8s_executor_instances"]}',
            f'--conf spark.kubernetes.namespace={app_args["k8s_namespace"]}',
            f'--conf spark.kubernetes.container.image={app_args["k8s_image_service"]}']

        spark_submit_aws = [
            '--packages org.apache.hadoop:hadoop-aws:3.3.4,com.amazonaws:aws-java-sdk-core:1.11.563,com.amazonaws:aws-java-sdk-s3:1.11.563',
            f'--conf spark.kubernetes.file.upload.path={app_args["k8s_upload_path"]}',
            f'--conf spark.kubernetes.driver.podTemplateFile={app_args["k8s_driver_podTemplateFile"]}',
            f'--conf spark.kubernetes.executor.podTemplateFile={app_args["k8s_executor_podTemplateFile"]}',
            '--conf spark.kubernetes.executor.deleteOnTermination=false',
            '--conf spark.kubernetes.container.imagePullPolicy=Always',
            '--conf spark.jars.ivy=/tmp/.ivy2',
            '--conf spark.hadoop.fs.s3a.aws.credentials.provider=org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider',
            '--conf spark.hadoop.fs.s3a.access.key="${AWS_ACCESS_KEY_ID}"',
            '--conf spark.hadoop.fs.s3a.secret.key="${AWS_SECRET_ACCESS_KEY}"',
            '--conf spark.hadoop.fs.s3a.session.token="${AWS_SESSION_TOKEN}"',
            '--conf spark.hadoop.fs.s3a.impl=org.apache.hadoop.fs.s3a.S3AFileSystem',
            f'--conf spark.hadoop.fs.s3a.endpoint=s3.{app_args["aws_region"]}.amazonaws.com',
            '--py-files tmp/files_to_ship/scripts.zip']

        spark_submit_docker_desktop = [
            '--conf spark.kubernetes.authenticate.driver.serviceAccountName=spark-service-account',
            '--conf spark.kubernetes.pyspark.pythonVersion=3',
            '--conf spark.pyspark.python=python3',
            '--conf spark.pyspark.driver.python=python3',
            '--conf spark.kubernetes.driver.volumes.hostPath.spark-local-dir.mount.path=/mnt/yaetos_jobs',
            '--conf spark.kubernetes.driver.volumes.hostPath.spark-local-dir.options.path=/Users/aprevot/Synced/github/code/code_perso/yaetos/',
            '--conf spark.kubernetes.executor.volumes.hostPath.spark-local-dir.mount.path=/mnt/yaetos_jobs',
            '--conf spark.kubernetes.executor.volumes.hostPath.spark-local-dir.options.path=/Users/aprevot/Synced/github/code/code_perso/yaetos/',
            '--conf spark.kubernetes.file.upload.path=file:///yaetos_jobs/tmp/files_to_ship/scripts.zip',
            '--py-files local:///mnt/yaetos_jobs/tmp/files_to_ship/scripts.zip']

        if app_args.get('k8s_mode') == 'k8s_aws':
            launcher = 'jobs/generic/launcher.py'
            spark_submit_conf_plateform = spark_submit_aws
        elif app_args.get('k8s_mode') == 'k8s_docker_desktop':
            launcher = 'local:///mnt/yaetos_jobs/jobs/generic/launcher.py'
            spark_submit_conf_plateform = spark_submit_docker_desktop
        else:
            raise Exception(f"k8s_mode not recognized, should be 'k8s_aws' or 'k8s_docker_desktop'. Set to {app_args.get('k8s_mode')}")

        spark_submit_conf_extra = app_args.get('spark_deploy_args', [])
        spark_submit_jobs = [
            launcher,
            f'--mode={app_args["mode"]}',  # need to make sure it uses a mode compatible with k8s
            '--deploy=none',
            '--storage=s3',
            f'--job_name={app_args["job_name"]}',
            '--runs_on=k8s']

        spark_submit_jobs_extra = app_args.get('spark_app_args', [])
        if app_args.get('dependencies'):
            spark_submit_jobs_extra += ['--dependencies']

        spark_submit = spark_submit_base \
            + spark_submit_conf_plateform \
            + spark_submit_conf_extra \
            + spark_submit_jobs \
            + spark_submit_jobs_extra
        return spark_submit

    @staticmethod
    def launch_spark_submit_k8s(self, cmdline):
        cmdline_str = " ".join(cmdline)
        logger.info(f'About to run spark submit command line (for reuse): {cmdline_str}')
        logger.info('About to run spark submit command line (for visual check): \n{}'.format(" \n".join(cmdline)))
        if not self.app_args.get('dry_run'):
            os.system(cmdline_str)

    @staticmethod
    def get_airflow_code(self, start_date, schedule):
        params = {
            'k8s_url': self.deploy_args.get('k8s_url'),
            'k8s_name': self.deploy_args.get('k8s_name'),
            'k8s_executor_instances': self.deploy_args.get('k8s_executor_instances'),
            'k8s_namespace': self.deploy_args.get('k8s_namespace'),
            'k8s_image_service': self.deploy_args.get('k8s_image_service'),
            'k8s_upload_path': self.deploy_args.get('k8s_upload_path'),
            'k8s_driver_podTemplateFile': self.deploy_args.get('k8s_driver_podTemplateFile'),
            'k8s_executor_podTemplateFile': self.deploy_args.get('k8s_executor_podTemplateFile'),
            'aws_region': self.s3_region,
            'k8s_podname': self.deploy_args.get('k8s_podname'),
            'k8s_airflow_spark_submit_yaml': self.deploy_args.get('k8s_airflow_spark_submit_yaml'),
            # airflow specific
            'dag_nameid': self.app_args['job_name'].replace("/", "-"),
            'start_date': start_date,
            'schedule': schedule,
            'emails': self.deploy_args.get('emails', '[]'),
        }
        param_extras = {key: self.deploy_args[key] for key in self.deploy_args if key.startswith('airflow.')}
        content = get_template_k8s(params, param_extras)
        return content



"""
Function to create a generic airflow dag for spark-submit to AWS EKS, generic enought to be used by any yaetos job.
"""
from textwrap import dedent


def get_template_k8s(params, param_extras):

    # Set extra params, params not available in template but overloadable
    lines = ''
    for item in param_extras.keys():
        entries = item.replace('airflow.', '').split('.')
        entries = '"]["'.join(entries)
        line = f'DAG_ARGS["{entries}"] = {param_extras[item]}\n' + ' ' * 4
        lines += line
    params['extras'] = lines

    template = """
    from airflow import DAG
    from airflow.providers.cncf.kubernetes.operators.spark_kubernetes import SparkKubernetesOperator
    from airflow.providers.cncf.kubernetes.sensors.spark_kubernetes import SparkKubernetesSensor
    from airflow.utils.dates import days_ago  # noqa: F401
    from datetime import timedelta
    import dateutil


    DAG_ARGS = {{
        'dag_id': '{dag_nameid}',
        'dagrun_timeout': timedelta(hours=2),
        'start_date': {start_date},
        'schedule': {schedule},
        'tags': ['emr'],
        'default_args': {{
            'owner': 'me',
            'depends_on_past': False,
            'email': {emails},
            'email_on_failure': False,
            'email_on_retry': False,
        }},
    }}
    {extras}

    with DAG(**DAG_ARGS) as dag:

        spark_submit = SparkKubernetesOperator(
            task_id='spark_submit_task',
            namespace='{k8s_namespace}',
            application_file='{k8s_airflow_spark_submit_yaml}',
            kubernetes_conn_id='k8s_default',
            do_xcom_push=True,
        )

        spark_sensor = SparkKubernetesSensor(
            task_id='watch_step',
            namespace='{k8s_namespace}',
            application_name="{{{{ task_instance.xcom_pull(task_ids='spark_submit_task')['metadata']['name'] }}}}",
            kubernetes_conn_id='k8s_default',
            poke_interval=60,
            timeout=600,
        )

        spark_submit >> spark_sensor
    """.format(**params)
    return dedent(template)



# replacement strings
WINDOWS_LINE_ENDING = b'\r\n'
UNIX_LINE_ENDING = b'\n'


def convert_to_linux_eol(path_in, path_out):
    """path_in can be absolute or relative, same for path_out.
    path_out can be the same as path_in, i.e. replacing in place."""
    with open(path_in, 'rb') as open_file:
        content = open_file.read()

    content = content.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

    with open(path_out, 'wb') as open_file:
        open_file.write(content)
    return True



import pandas as pd
from yaetos.etl_utils import Path_Handler
from yaetos.db_utils import pdf_to_sdf
from yaetos.logger import setup_logging
logger = setup_logging('Job')


def load_excel(jargs, input_name, output_types, sc, sc_sql, **xls_args):
    pdf = load_excel_pandas(jargs, input_name, **xls_args)
    sdf = pdf_to_sdf(pdf, output_types, sc, sc_sql)
    return sdf


def load_excel_pandas(jargs, input_name, **xls_args):

    path = jargs.inputs[input_name]['path']
    path = path.replace('s3://', 's3a://') if 'dev_local' in jargs.mode.split(',') else path
    logger.info("Input '{}' to be loaded from files '{}'.".format(input_name, path))
    path = Path_Handler(path, jargs.base_path).expand_latest()
    logger.info("Input '{}' loaded from files '{}'.".format(input_name, path))

    pdf = pd.read_excel(io=path, engine='openpyxl', **xls_args)
    return pdf



from yaetos.logger import setup_logging
logger = setup_logging('Deploy')


def terminate(error_message=None):
    """
    Method to exit the Python script. It will log the given message and then exit().
    :param error_message:
    """
    if error_message:
        logger.error(error_message)
    logger.critical('The script is now terminating')
    exit()



from pyspark import SparkContext
from pyspark.sql import SQLContext


def query_sparksql_local(query_str, fnames_in=None, dfs_in=None, **kwargs):
    """Slower than query_pandasql.py but supports windowing functions."""
    sc = SparkContext("local", "local_run")
    sc_sql = SQLContext(sc)
    run_sql = sc_sql.sql

    if fnames_in:
        for name, fname in fnames_in.items():
            reader = sc_sql.read.option("header", "true")
            if kwargs.get('delimiter'):
                reader = reader.option("delimiter", kwargs['delimiter'])
            # if kwargs.get('quoting'):  # function tried (using jobisjob_apipull_missing_advertisers_with_lkin_process.py) but didn't work. TODO: get it working.
            #     reader = reader.option('quote', '"').option('escape', '"')
            df = reader.csv(fname)
            df.createOrReplaceTempView(name)
            print("Input {}, data types: {}".format(name, [(fd.name, fd.dataType) for fd in df.schema.fields]))

    if dfs_in:
        # Input dfs may require enforced standard types in pandas dfs (no "object"), so they are compatible with spark dfs.
        for name, df in dfs_in.items():
            df = sc_sql.createDataFrame(df)
            df.createOrReplaceTempView(name)
            print("Input {}, data types: {}".format(name, [(fd.name, fd.dataType) for fd in df.schema.fields]))

    df_out = run_sql(query_str)
    df_out.cache()
    print('Sample output: {}'.format(df_out.show()))  # TODO: rely on upstream functions arg 'show' when done before dropping.
    pdf = df_out.toPandas()
    sc.stop()
    return pdf



"""
Job to showcase loading file as pandas dataframe instead of spark dataframe.
This allows for faster run for small datasets but looses some of the benefits of spark dataframes (support for SQL, better field type management, etc.).
Transformation is the same as ex1_sql_job.sql
"""
from yaetos.etl_utils import ETL_Base, Commandliner
import pandas as pd
import numpy as np


class Job(ETL_Base):
    def transform(self, some_events, other_events):
        some_events = some_events[:1000]  # to speed up since it is an example job
        df1 = some_events[some_events.apply(lambda row: row['action'] == 'searchResultPage' and float(row['n_results']) > 0, axis=1)]
        df2 = pd.merge(left=df1, right=other_events, how='inner', left_on='session_id', right_on='session_id', indicator=True, suffixes=('_1', '_2'))
        df3 = df2.groupby(by=['session_id']).agg({'_merge': np.count_nonzero})
        df3.rename(columns={'_merge': 'count_events'}, inplace=True)
        df3.sort_values('count_events', ascending=False, inplace=True)
        df3.reset_index(drop=False, inplace=True)
        self.logger.info('Post filter length: {}'.format(len(df1)))
        return df3


if __name__ == "__main__":
    args = {'job_param_file': 'conf/jobs_metadata.yml'}
    Commandliner(Job, **args)



"""Same as ex1_sql_job.sql but allows access to spark for more complex ops (not used here but in ex2_frameworked_job.py)."""
from yaetos.etl_utils import ETL_Base, Commandliner


class Job(ETL_Base):
    def transform(self, some_events, other_events):
        df = self.query("""
            SELECT se.session_id, count(*) as count_events
            FROM some_events se
            JOIN other_events oe on se.session_id=oe.session_id
            WHERE se.action='searchResultPage' and se.n_results>0
            group by se.session_id
            order by count(*) desc
            """)
        return df


if __name__ == "__main__":
    args = {'job_param_file': 'conf/jobs_metadata.yml'}
    Commandliner(Job, **args)



"""Helper functions for redshift."""

from yaetos.logger import setup_logging
logger = setup_logging('Redshift')


def create_table(df, connection_profile, name_tb, schema, creds_or_file, is_incremental, s3_tmp_dir, spark_version):
    """
    Creates table in redshift, full drop or incremental drop, using spark connector. Implies pushing data to S3 first.
    """
    load_type = 'overwrite' if not is_incremental else 'append'
    db = creds_or_file[connection_profile]
    url = 'jdbc:redshift://{host}:{port}/{service}'.format(host=db['host'], port=db['port'], service=db['service'])
    dbtable = '{}.{}'.format(schema, name_tb)

    logger.info('Sending table "{}" to redshift in schema "{}", load type "{}", size "{}".'.format(name_tb, schema, load_type, df.count()))
    format = 'io.github.spark_redshift_community.spark.redshift' if spark_version == '3.1' else 'com.databricks.spark.redshift'  # else implies spark_version == '2.4'

    df.write \
        .format(format) \
        .option("tempdir", s3_tmp_dir) \
        .option("url", url) \
        .option("forward_spark_s3_credentials", "true") \
        .option("user", db['user']) \
        .option("password", db['password']) \
        .option("dbtable", dbtable) \
        .mode(load_type) \
        .save()

    logger.info("Copied table to redshift '{}.{}', using connection profile '{}'".format(schema, name_tb, connection_profile))



from yaetos.etl_utils import ETL_Base, Commandliner
from yaetos.libs.analysis_toolkit.compare_pandas_dfs import compare_dfs_fuzzy, compare_dfs_exact
import yaetos.pandas_utils as pu
import pandas as pd


class Job(ETL_Base):

    def transform(self, tableA, tableB):
        pksA = self.jargs.inputs['tableA']['pk'] if self.jargs.inputs['tableA'].get('pk') else list(tableA.columns)
        pksB = self.jargs.inputs['tableB']['pk'] if self.jargs.inputs['tableB'].get('pk') else list(tableB.columns)
        compareA = self.jargs.inputs['tableA']['compare'] if self.jargs.inputs['tableA'].get('compare') else list(set(tableA.columns) - set(pksA))
        compareB = self.jargs.inputs['tableB']['compare'] if self.jargs.inputs['tableB'].get('compare') else list(set(tableB.columns) - set(pksB))

        return compare_dfs(tableA, tableB, pksA, pksB, compareA, compareB)


def compare_dfs(tableA, tableB, pksA, pksB, compareA, compareB):
    # Comparing columns
    tableA.reset_index(drop=True, inplace=True)
    tableB.reset_index(drop=True, inplace=True)

    print('Stats for both tables (numerical columns only)')
    print('tableA: ', tableA.describe())
    print('tableB: ', tableB.describe())

    # Comparing columns
    print('Comparing columns')
    diff_columns = list(set(tableA.columns) ^ set(tableB.columns))
    common_columns = list(set(tableA.columns) & set(tableB.columns))

    if diff_columns:
        print('datasets are different')
        print('The deltas in columns are: ', diff_columns)
        print('The columns in common are: ', common_columns)
        print('Rest of the comparison will be based on columns in common')
        # tableA = tableA.drop_duplicates()
        # tableB = tableB.drop_duplicates()
    else:
        print(f'datasets have same columns, i.e. {common_columns}')
    print(f'TableA columns {tableA.columns}')
    print(f'TableB columns {tableB.columns}')
    tableA = tableA[common_columns]  # applied even if columns are all the same to make sure they are in the same order (impt for checks below)
    tableB = tableB[common_columns]

    # Comparing datasets length
    print('Comparing row count')
    diff_row_count = len(tableA) != len(tableB)
    if diff_row_count:
        print('datasets have different row count')
        print('Length TableA: ', len(tableA))
        print('Length TableB: ', len(tableB))

    # Comparing dataset content, exact
    print('Comparing dataset - exact content match')
    is_identical = compare_dfs_exact(tableA, tableB)
    if is_identical:
        message = 'datasets are identical' if not diff_columns else 'datasets (with columns in common) are identical'
        print(message)
        return pd.DataFrame()

    # Comparing dataset content, fuzzy
    print('Comparing dataset - fuzzy content match (match specific columns, and check float deltas within threshold)')
    strip = False
    filter_deltas = False

    if not pu.check_pk(tableA, pksA):
        print('The chosen PKs are not actual PKs (i.e. not unique). Retry with modified PK inputs.')
        return pd.DataFrame()
    elif not pu.check_pk(tableB, pksB):
        print('The chosen PKs are not actual PKs (i.e. not unique). Retry with modified PK inputs.')
        return pd.DataFrame()

    print('About to compare, column by column.')
    df_out = compare_dfs_fuzzy(tableA, pksA, compareA, tableB, pksB, compareB, strip, filter_deltas, threshold=0.01)
    print('Finishing compare, column by column.')

    return df_out


if __name__ == "__main__":
    args = {'job_param_file': 'conf/jobs_metadata.yml'}
    Commandliner(Job, **args)



#!/usr/bin/env python
"""
Script to setup commandline for yaetos when installed with pip install.

Usage:
 * pip install yaetos
 * cd /path/to/an/empty/folder/that/will/contain/pipeline/code
 * yaetos --help   # to get the options

For dev:
 * To use lib without publishing it: "cd path/to/repo/yeatos; pip install ." # TODO: check why changes are not always picked up.
 * Alternative if last step doesn't work (due to unusual python or pip setup): "python -c 'from yaetos.scripts.yaetos_cmdline import YaetosCmds; YaetosCmds()'"
 * This feeds into setup.py
"""

import os
import stat
from pathlib import Path as Pt
from shutil import copyfile
import yaetos
import argparse
import sys
import subprocess


class YaetosCmds(object):
    # Source: https://chase-seibert.github.io/blog/2014/03/21/python-multilevel-argparse.html

    usage_setup = "Setup yaetos folders and files in current folder. This will overwrite files if any."
    usage_docker_bash = "Launching docker container to run jobs from bash."
    usage_docker_jupyter = "Launching docker container to run jobs from jupyter notebook."
    usage_run_dockerized = "Run job through docker. Running 'yaetos run_dockerized some/job.py --some=arg' is the same as running 'yaetos usage_docker_bash; python some/job.py --some=arg'"
    usage_run = "Run job in terminal. Running 'yaetos run some/job.py --some=arg' is the same as running 'python some/job.py --some=arg'"
    usage_panel = "Run panel commandlines to create dashboards (experimental). Running 'yaetos panel some_args' is the same as running 'panel some_args'"

    usage = f'''
    yaetos <command> [<args>]

    Yaetos top level commands are:
    setup                : {usage_setup}
    launch_docker_bash   : {usage_docker_bash}
    launch_docker_jupyter: {usage_docker_jupyter}
    run_dockerized       : {usage_run_dockerized}
    run                  : {usage_run}
    panel                : {usage_panel}

    For more details, see:
    - https://yaetos.readthedocs.io/en/latest/
    '''

    def __init__(self):
        parser = argparse.ArgumentParser(
            description='Yeatos command lines',
            usage=self.usage)
        parser.add_argument('command', help='Subcommand to run')
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print('Unrecognized command')
            parser.print_help()
            exit(1)
        getattr(self, args.command)()  # dispatching according to command line.

    def setup(self):
        parser = argparse.ArgumentParser(description=self.usage_setup)
        parser.add_argument('--set_github', action='store_true', help="Identify registry job to use.")
        parser.add_argument('--project_name', help="Set the name of the project, i.e. the name of the folder that will contain the job files.")
        args = parser.parse_args(sys.argv[2:])  # ignoring first 2 args (i.e. "yeatos setup")
        setup_env(args)

    def launch_docker_bash(self):
        argparse.ArgumentParser(description=self.usage_docker_bash)
        # parser.add_argument('--no_aws', action='store_true')  # TODO: implement
        out = subprocess.call("./launch_env.sh 1", shell=True)  # TODO: make it work with better: subprocess.call(["./launch_env.sh", '1']) # TODO: make it cross plateform (check using "sh launch_env.sh 1" for windows)
        self.print_error(out)

    def launch_docker_jupyter(self):
        argparse.ArgumentParser(description=self.usage_docker_jupyter)
        out = subprocess.call("./launch_env.sh 2", shell=True)
        self.print_error(out)

    def run_dockerized(self):
        parser = argparse.ArgumentParser(description=self.usage_run_dockerized)
        ignored, cmd_unknown_args = parser.parse_known_args()
        cmd_str = 'python ' + ' '.join(cmd_unknown_args[1:])
        cmd_delegated = "./launch_env.sh 3 " + cmd_str
        out = subprocess.call(cmd_delegated, shell=True)
        self.print_error(out)

    def run(self):
        parser = argparse.ArgumentParser(description=self.usage_run)
        ignored, cmd_unknown_args = parser.parse_known_args()
        cmd_str = 'python ' + ' '.join(cmd_unknown_args[1:])
        cmd_delegated = "./launch_env.sh 4 " + cmd_str
        out = subprocess.call(cmd_delegated, shell=True)
        self.print_error(out)

    def panel(self):
        parser = argparse.ArgumentParser(description=self.usage_panel)
        ignored, cmd_unknown_args = parser.parse_known_args()
        cmd_str = ' '.join(cmd_unknown_args[1:])
        cmd_delegated = "./launch_env.sh 4 " + cmd_str
        out = subprocess.call(cmd_delegated, shell=True)
        self.print_error(out)

    def print_error(self, out):
        if out == 127:  # 127 associated to error "/bin/sh: ./launch_env.sh: No such file or directory"
            print('Error: this command needs to be run from the project root, where launch_env.sh is, or have --path pointing to that folder.')


def main():
    # Necessary for setup.py entry_points to point to a function instead of a class (to avoid printing class repr in terminal)
    YaetosCmds()


def setup_env(args):
    cwd = os.getcwd()
    if args.project_name:
        os.makedirs(Pt(args.project_name), exist_ok=True)
        os.chdir(Pt(args.project_name))
        cwd = os.getcwd()
        print(f'Created the folder "{args.project_name}"')

    paths = yaetos.__path__
    package_path = paths[0]
    if len(paths) > 1:
        print(f'Yeatos python package found in several locations. The script will use this one: {package_path}')

    # Empty folders necessary for later.
    os.makedirs(Pt('tmp/files_to_ship/'), exist_ok=True)
    os.makedirs(Pt('data/'), exist_ok=True)
    # TODO: make code above and below compatible with windows OS (no cmd line, no linux only paths).

    # Root folder files
    copyfile(Pt(f'{package_path}/scripts/copy/Dockerfile_external'), Pt(f'{cwd}/Dockerfile'))
    copyfile(Pt(f'{package_path}/scripts/copy/launch_env_external.sh'), Pt(f'{cwd}/launch_env.sh'))
    copyfile(Pt(f'{package_path}/scripts/copy/_gitignore_external'), Pt(f'{cwd}/.gitignore'))
    os.chmod(Pt(f'{cwd}/launch_env.sh'), stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)  # equ to 0o755

    # Conf
    os.makedirs(Pt('conf/'), exist_ok=True)
    copyfile(Pt(f'{package_path}/scripts/copy/aws_config.cfg.example'), Pt(f'{cwd}/conf/aws_config.cfg'))
    copyfile(Pt(f'{package_path}/scripts/copy/jobs_metadata_external.yml'), Pt(f'{cwd}/conf/jobs_metadata.yml'))
    copyfile(Pt(f'{package_path}/scripts/copy/connections.cfg.example'), Pt(f'{cwd}/conf/connections.cfg'))
    copyfile(Pt(f'{package_path}/scripts/copy/requirements_extra.txt'), Pt(f'{cwd}/conf/requirements_extra.txt'))

    # Generic jobs
    os.makedirs(Pt('jobs/generic/'), exist_ok=True)
    copyfile(Pt(f'{package_path}/libs/generic_jobs/copy_job.py'), Pt(f'{cwd}/jobs/generic/copy_job.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/copy_raw_job.py'), Pt(f'{cwd}/jobs/generic/copy_raw_job.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/deployer.py'), Pt(f'{cwd}/jobs/generic/deployer.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/dummy_job.py'), Pt(f'{cwd}/jobs/generic/dummy_job.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/launcher.py'), Pt(f'{cwd}/jobs/generic/launcher.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/sql_spark_job.py'), Pt(f'{cwd}/jobs/generic/sql_spark_job.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/sql_pandas_job.py'), Pt(f'{cwd}/jobs/generic/sql_pandas_job.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/compare_job.py'), Pt(f'{cwd}/jobs/generic/compare_job.py'))
    copyfile(Pt(f'{package_path}/libs/generic_jobs/list_files_job.py'), Pt(f'{cwd}/jobs/generic/list_files_job.py'))

    # Sample jobs
    os.makedirs(Pt('jobs/examples/'), exist_ok=True)
    copyfile(Pt(f'{package_path}/scripts/copy/ex0_extraction_job.py'), Pt(f'{cwd}/jobs/examples/ex0_extraction_job.py'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_pandas_api_job.py'), Pt(f'{cwd}/jobs/examples/ex1_pandas_api_job.py'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_sql_job.sql'), Pt(f'{cwd}/jobs/examples/ex1_sql_job.sql'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_spark_api_job.py'), Pt(f'{cwd}/jobs/examples/ex1_spark_api_job.py'))

    # Sample jobs tests
    os.makedirs(Pt('tests/jobs/examples/'), exist_ok=True)
    copyfile(Pt(f'{package_path}/scripts/copy/conftest.py'), Pt(f'{cwd}/tests/conftest.py'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_pandas_api_job_test.py'), Pt(f'{cwd}/tests/jobs/examples/ex1_pandas_api_job_test.py'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_spark_api_job_test.py'), Pt(f'{cwd}/tests/jobs/examples/ex1_spark_api_job_test.py'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_sql_pandas_job_test.py'), Pt(f'{cwd}/tests/jobs/examples/ex1_sql_pandas_job_test.py'))
    copyfile(Pt(f'{package_path}/scripts/copy/ex1_sql_spark_job_test.py'), Pt(f'{cwd}/tests/jobs/examples/ex1_sql_spark_job_test.py'))

    # TODO: add setup awscli or make sure it is there.

    # setup github CI
    if args.set_github:
        os.makedirs(Pt('.github/workflows/'), exist_ok=True)
        copyfile(Pt(f'{package_path}/scripts/github_pythonapp.yml'), Pt(f'{cwd}/.github/workflows/pythonapp.yml'))

    print("""Filled the folder with:
     * sample jobs, ready for local execution,
     * a docker environment setup with spark,
     * a configuration file for cloud execution,
     * job unit-tests.""")



"""Helper functions for kafka.
"""

import json
import jsonschema
import kafka
import requests
from yaetos.logger import setup_logging
logger = setup_logging('Kafka_push')


class KafkaProducer(object):

    def __init__(self, broker_address, topic, send_timeout, check_schema=False, schema_uri=None, connect_kafka=True):
        # TODO: add schema validation to only do message validation in the later stage (send()) to avoid validating schema for every record.
        self.__send_timeout = send_timeout
        self.__TOPIC = topic
        self.check_schema = check_schema
        if check_schema:
            self.schema_init(schema_uri)
        if connect_kafka:
            self.__producer = kafka.KafkaProducer(bootstrap_servers=[broker_address])
            logger.info("Producer connection started")

    def schema_init(self, schema_uri):
        self.__SCHEMA_URI = schema_uri
        response = requests.get(self.__SCHEMA_URI)
        schema = response.json()
        logger.info("Producer schema uri to pull schema data: {}".format(schema_uri))
        logger.info("Producer schema data pulled from schema service: {}.".format(schema))
        schema["$schema"] = self.__SCHEMA_URI  # DIRTY FIX
        self.__schema = schema

    def build_message(self, **rec):
        """To be overriden by parent class"""
        raise NotImplementedError

    def send(self, rec, **kwargs):
        message = self.build_message(rec, **kwargs)  # TODO: remove this step from this function, or rename send() to build_and_send()

        to_send = True
        if self.check_schema:
            to_send = self.validate(message)

        # Actual send part
        if to_send:
            msg_in = bytes(bytearray(json.dumps(message), "UTF-8"))
            msg_out = self.__producer.send(self.__TOPIC, msg_in)

            # Post send
            msg_out.get(timeout=self.__send_timeout)
            return msg_out

    def validate(self, message):
        try:
            jsonschema.validate(instance=message, schema=self.__schema)
            return True
        except jsonschema.exceptions.ValidationError as e:
            logger.error(u"Validation error pre kafka send: {}.".format(str(e)))
            return False

    def flush(self):
        self.__producer.flush()

    def close(self):
        self.__producer.close()
        logger.info("Producer connection closed")

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        self.close()


if __name__ == '__main__':
    pass



import logging


def setup_logging(logger_name, default_level=logging.INFO):
    """
    Setup logging configuration
    Use in scripts: logger = setup_logging('Job')
    """
    format = "|%(asctime)s:%(levelname)s:%(name)s|%(message)s"  # Default:   format = "%(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=default_level, format=format, datefmt="%H:%M:%S")
    logging.getLogger('Deploy').setLevel(logging.INFO)
    logging.getLogger('botocore.credentials').setLevel(logging.CRITICAL)
    logging.getLogger('boto3.resources.action').setLevel(logging.CRITICAL)
    return logging.getLogger(logger_name)



from yaetos.etl_utils import ETL_Base, Commandliner, Cred_Ops_Dispatcher
from yaetos.db_utils import pdf_to_sdf
from libs.python_db_connectors.query_mysql import query as query_mysql


class Mysql_Job(ETL_Base):

    def query_mysql(self, query_str):
        """ Requires OUTPUT_TYPES to be specified in class.
        and mysql connection to be put in yml file in line like "api_inputs: {'api_creds': 'mysql_creds_section'}"
        output spark df."""
        query_str = query_str.replace('%', '%%')  # replace necessary for pymysql parser
        self.logger.info('Query string:\n' + query_str)
        creds = Cred_Ops_Dispatcher().retrieve_secrets(self.jargs.storage, creds=self.jargs.connection_file)
        creds_section = self.jargs.yml_args['api_inputs']['api_creds']
        self.logger.info('The query is using the credential section:' + creds_section)
        pdf = query_mysql(query_str, db=creds_section, creds_or_file=creds)
        sdf = pdf_to_sdf(pdf, self.OUTPUT_TYPES, self.sc, self.sc_sql)
        return sdf


if __name__ == "__main__":
    Commandliner(Mysql_Job)
